import java.awt.Point;


/** A Piece represents an instance of a puzzle piece, with a location. */

public class Piece
{
    static final int DX[] = { 1, 0, -1, 0 };            // directional offsets
    static final int DY[] = { 0, 1, 0, -1 };

    Point   where;              // where is the upper left corner of this piece?
    int     shapeIndex;         // what shape is it?  (index into shapes array in
                                // input spec)
    
    
    /** the input specifications */
    
    public static InputSpec spec;
    
    
    /** Make a new piece of a given shape, in a given place. */

    public Piece(int whichShape, Point w)
    {
        shapeIndex = whichShape;
        where = w;
    }
    
    
    /** Make a copy of a piece. */
     
    public Piece(Piece piece)
    {
        where = new Point(piece.where.x, piece.where.y);
        shapeIndex = piece.shapeIndex;
    }
    
    /** Is this the goal piece, in the goal position? */

    public boolean solved()
    {
        return isGoal() && where.equals(spec.goalLocation);
    }
    
    
    /** Is this the goal piece? */

    public boolean isGoal()
    {
        return shapeIndex == spec.goalIndex;
    }
    

    /** Does this piece occupy the spot (x, y)? */

    public boolean occupies(int x, int y)
    {
        x -= where.x;
        y -= where.y;
        
        return shape().occupies(x, y);
    }
    
    
    /** What is this piece's shape? */
    
    public Shape shape()
    {
        return spec.pieceShapes[shapeIndex];
    }
    
    
    /** Can this piece slide in the specified direction, in the given configuration?
        We must make sure that the squares the piece would slide into are vacant.
        Those are the ones that are in the given direction from the squares the
        piece occupies, except for those such squares that the piece itself
        currently occupies. */

    public boolean canSlide(int dir, Configuration config)
    {
        for (int x = where.x; x <= where.x + shape().width(); ++x)
            for (int y = where.y; y <= where.y + shape().height(); ++y)
                if (occupies(x, y) && !occupies(x + DX[dir], y + DY[dir]) && config.occupied(x + DX[dir], y + DY[dir]))
                    return false;
                    
        return true;
    }
    
    
    /**  Shift the piece. */

    public void shift(int dir)
    {
        where.translate(DX[dir], DY[dir]);
    }
    
    
    /** Print out the piece info. */

    public void print()
    {
        System.out.println(shapeIndex + " " + where.x + " " + where.y);
    }
    

    /** Is this piece the same shape as and in the same location as another piece? */

    public boolean equals(Piece piece)
    {
        return shapeIndex == piece.shapeIndex && where.equals(piece.where);
    }
}
